package com.flow.listener;

import com.baomidou.mybatisplus.core.toolkit.StringUtils;
import com.flow.common.oauth2.utils.SecurityContextUtil;
import com.flow.constant.NodeTypeConstant;
import com.flow.entity.*;
import com.flow.entity.node.ApprovalNode;
import com.flow.entity.node.StartNode;
import com.flow.entity.settings.CancelConfig;
import com.flow.entity.settings.Settings;
import com.flow.enums.ApprovalNobodyEnum;
import com.flow.enums.AssigneeTypeEnum;
import com.flow.enums.NotifyEnum;
import com.flow.enums.ProcessStatus;
import com.flow.service.*;
import lombok.extern.slf4j.Slf4j;
import org.flowable.common.engine.api.delegate.event.FlowableEngineEntityEvent;
import org.flowable.engine.TaskService;
import org.flowable.engine.delegate.DelegateExecution;
import org.flowable.engine.delegate.event.AbstractFlowableEngineEventListener;
import org.flowable.engine.delegate.event.FlowableCancelledEvent;
import org.flowable.engine.delegate.event.FlowableProcessStartedEvent;
import org.flowable.engine.delegate.event.impl.FlowableProcessCancelledEventImpl;
import org.flowable.engine.impl.persistence.entity.ExecutionEntityImpl;
import org.flowable.task.service.impl.persistence.entity.TaskEntityImpl;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;
import org.springframework.transaction.support.TransactionSynchronization;
import org.springframework.transaction.support.TransactionSynchronizationManager;

import java.time.LocalDateTime;
import java.util.Map;
import java.util.Objects;

@Component
@Slf4j
public class GlobalActivityEventListener extends AbstractFlowableEngineEventListener {
    @Autowired
    private FlowDefineService flowDefineService;
    @Autowired
    private FlowTaskService flowTaskService;
    @Autowired
    private TaskService taskService;
    @Autowired
    private FlowActivityService flowActivityService;
    @Autowired
    private NotifyService notifyService;
    @Autowired
    private FlowInstanceService flowInstanceService;

    // 任务创建监听器
    @Override
    protected void taskCreated(FlowableEngineEntityEvent event) {
        TaskEntityImpl entity = (TaskEntityImpl) event.getEntity();

        // 更新任务状态
        FlowTask flowTask = FlowTask.buildStatus(entity.getId(), ProcessStatus.RUNNING);
        flowTaskService.save(flowTask);

        // 记录日志
        FlowActivity flowActivity = flowActivityService.build(entity);
        flowActivity.setNodeType(NodeTypeConstant.APPROVAL);
        flowActivity.setStatus(ProcessStatus.RUNNING);
        flowActivityService.save(flowActivity);


        FlowDefine define = flowDefineService.getDefine(entity.getProcessDefinitionId());
        ApprovalNode approvalNode = define.getNode(entity.getTaskDefinitionKey(), ApprovalNode.class);
        if(Objects.nonNull(approvalNode)){
            AssigneeTypeEnum assigneeType = approvalNode.getAssigneeType();
            ApprovalNobodyEnum nobody = approvalNode.getNobody();

            // 自动通过
            if (assigneeType == AssigneeTypeEnum.AUTO_PASS) {
                flowTaskService.autoComplete(entity.getId());
                return;
            }

            // 自动拒绝
            if (assigneeType == AssigneeTypeEnum.AUTO_REFUSE) {
                flowTaskService.autoRefuse(entity.getId());
                return;
            }
            // 审批人为空方式
            if (StringUtils.isBlank(entity.getAssignee()) || "empty".equals(entity.getAssignee())) {
                entity.setAssignee("admin");
                taskService.setAssignee(entity.getId(), "admin");
                if (nobody == ApprovalNobodyEnum.PASS) {
                    flowTaskService.autoComplete(entity.getId(), " 审批人为空");
                } else if (nobody == ApprovalNobodyEnum.REFUSE) {
                    flowTaskService.autoRefuse(entity.getId(), " 审批人为空");
                }
                return;
            }
        }

        // 消息通知
        TransactionSynchronizationManager.registerSynchronization(new TransactionSynchronization() {
            @Override
            public void afterCommit() {
                Notify notify = new Notify();
                notify.setSubject("新的审批任务");
                notify.setContent(String.format("您有一条待处理的审批任务【%s】，请及时处理。", entity.getName()));
                notify.setSender("admin");
                notify.setType(NotifyEnum.TODO);
                notify.setReceiver(entity.getAssignee());
                notify.setReceivingTime(LocalDateTime.now());
                notify.setUrl("/flow/todo");
                notifyService.notify(notify);
            }
        });
    }

    // 任务完成监听器
    @Override
    protected void taskCompleted(FlowableEngineEntityEvent event) {
        TaskEntityImpl entity = (TaskEntityImpl) event.getEntity();
        log.info("任务完成监听器");
        // 更新日志
        flowActivityService.lambdaUpdate()
                .set(FlowActivity::getEndTime, LocalDateTime.now())
                .set(FlowActivity::getStatus, ProcessStatus.COMPLETED)
                .set(FlowActivity::getUserId, entity.getAssignee())
                .eq(FlowActivity::getTaskId, entity.getId())
                .update();
    }


    // 流程创建监听器
    @Override
    protected void processStarted(FlowableProcessStartedEvent event) {
        ExecutionEntityImpl entity = (ExecutionEntityImpl) event.getEntity();
        // 更新流程实例状态
        FlowDefine define = flowDefineService.getDefine(entity.getProcessDefinitionId());
        FlowInstance flowInstance = FlowInstance.buildStatus(entity.getProcessInstanceId(), ProcessStatus.RUNNING);
        Settings settings = define.getSettings();
        CancelConfig cancel = settings.getCancel();
        if (cancel.getEnable()) {
            flowInstance.setCancelDays(cancel.getDays());
        } else {
            flowInstance.setCancelDays(0);
        }
        flowInstanceService.save(flowInstance);
        // 记录日志
        FlowActivity flowActivity = flowActivityService.build(entity);
        flowActivity.setParentId(entity.getActivityId());
        flowActivity.setNodeType(NodeTypeConstant.START);
        flowActivity.setEndTime(LocalDateTime.now());
        flowActivity.setUserId(SecurityContextUtil.getUserId());
        flowActivity.setStatus(ProcessStatus.COMPLETED);
        flowActivityService.save(flowActivity);
    }

    // 流程完成监听器
    @Override
    protected void processCompleted(FlowableEngineEntityEvent event) {
        ExecutionEntityImpl entity = (ExecutionEntityImpl) event.getEntity();
        Map<String, Object> variables = entity.getVariables();
        // 保存流程变量
        FlowVariable flowVariable = new FlowVariable();
        flowVariable.setId(entity.getProcessInstanceId());
        flowVariable.setVariables(variables);
        // 更新流程实例状态
        FlowInstance flowInstance = FlowInstance.buildStatus(entity.getProcessInstanceId(), ProcessStatus.COMPLETED);
        flowInstanceService.updateById(flowInstance);
        // 记录日志
        FlowActivity flowActivity = flowActivityService.build(entity);
        flowActivity.setNodeType(NodeTypeConstant.END);
        flowActivity.setNodeId("end");
        flowActivity.setParentId("end");
        flowActivity.setNodeName("结束");
        flowActivity.setStartTime(LocalDateTime.now());
        flowActivity.setEndTime(LocalDateTime.now());
        flowActivity.setStatus(ProcessStatus.COMPLETED);
        flowActivityService.save(flowActivity);


        // 消息通知
        TransactionSynchronizationManager.registerSynchronization(new TransactionSynchronization() {
            @Override
            public void afterCommit() {
                Notify notify = new Notify();
                notify.setSubject("申请已通过");
                notify.setContent(String.format("您申请的流程《%s》已通过", entity.getName()));
                notify.setSender("admin");
                notify.setType(NotifyEnum.SYSTEM);
                notify.setReceiver(entity.getStartUserId());
                notify.setReceivingTime(LocalDateTime.now());
                notify.setUrl("/flow/instance");
                notifyService.notify(notify);
            }
        });
    }

    // 流程取消监听器
    @Override
    protected void processCancelled(FlowableCancelledEvent event) {
        FlowableProcessCancelledEventImpl flowableProcessCancelledEvent = (FlowableProcessCancelledEventImpl) event;
        DelegateExecution execution = flowableProcessCancelledEvent.getExecution();
        Map<String, Object> variables = execution.getVariables();
        FlowVariable flowVariable = new FlowVariable();
        flowVariable.setId(flowableProcessCancelledEvent.getProcessInstanceId());
        flowVariable.setVariables(variables);
    }

}
